import mongoChangeStream from 'mongochangestream';
import _ from 'lodash/fp.js';
import { renameId, setDefaults, sumByRowcount } from './util.js';
import { convertSchema } from './convertSchema.js';
import _debug from 'debug';
const debug = _debug('mongo2crate:sync');
export const initSync = (redis, collection, crate, options = {}) => {
    const mapper = options.mapper || renameId;
    const schemaName = options.schemaName || 'doc';
    const tableName = options.tableName || collection.collectionName.toLowerCase();
    const qualifiedName = `"${schemaName}"."${tableName}"`;
    // Initialize sync
    const sync = mongoChangeStream.initSync(redis, collection, options);
    // Use emitter from mongochangestream
    const emitter = sync.emitter;
    const emit = (event, data) => {
        emitter.emit(event, { type: event, ...data });
    };
    const createTableFromSchema = async (jsonSchema, options = {}) => {
        const createTableStmt = convertSchema(jsonSchema, qualifiedName, {
            omit: options.omit,
            ...options,
        });
        return crate.query(createTableStmt);
    };
    const handleResult = (result) => {
        debug('Change stream result %O', result);
        if ('rowcount' in result) {
            emit('process', { success: result.rowcount, changeStream: true });
        }
        else {
            emit('error', { error: result, changeStream: true });
        }
    };
    /**
     * Process a change stream event.
     */
    const processRecord = async (doc) => {
        try {
            if (doc.operationType === 'insert') {
                const document = mapper(doc.fullDocument);
                const result = await crate.insert(qualifiedName, document);
                handleResult(result);
            }
            else if (doc.operationType === 'update') {
                const document = doc.fullDocument ? mapper(doc.fullDocument) : {};
                const { updatedFields, removedFields } = doc.updateDescription;
                const removed = removedFields && setDefaults(removedFields, null);
                const update = mapper({ ...updatedFields, ...removed });
                if (_.size(update)) {
                    const result = await crate.upsert(qualifiedName, document, update);
                    handleResult(result);
                }
            }
            else if (doc.operationType === 'replace') {
                const id = doc.documentKey._id.toString();
                // Delete
                await crate.deleteById(qualifiedName, id);
                // Insert
                const document = mapper(doc.fullDocument);
                const result = await crate.insert(qualifiedName, document);
                handleResult(result);
            }
            else if (doc.operationType === 'delete') {
                const id = doc.documentKey._id.toString();
                const result = await crate.deleteById(qualifiedName, id);
                handleResult(result);
            }
        }
        catch (e) {
            emit('error', { error: e, changeStream: true });
        }
    };
    /**
     * Process initial scan documents.
     */
    const processRecords = async (docs) => {
        try {
            const documents = docs.map(({ fullDocument }) => mapper(fullDocument));
            const result = await crate.bulkInsert(qualifiedName, documents);
            debug('Initial scan result %O', result);
            if ('results' in result) {
                const numInserted = sumByRowcount(1)(result.results);
                const numFailed = sumByRowcount(-2)(result.results);
                emit('process', {
                    success: numInserted,
                    fail: numFailed,
                    initialScan: true,
                });
            }
            if ('error' in result) {
                emit('error', { error: result, initialScan: true });
            }
        }
        catch (e) {
            emit('error', { error: e, initialScan: true });
        }
    };
    const processChangeStream = (options) => sync.processChangeStream(processRecord, options);
    const runInitialScan = (options) => sync.runInitialScan(processRecords, options);
    return {
        ...sync,
        /**
         * Process MongoDB change stream for the given collection.
         */
        processChangeStream,
        /**
         * Run initial collection scan. `options.batchSize` defaults to 500.
         * Sorting defaults to `_id`.
         */
        runInitialScan,
        /**
         * Convert the given JSON schema to CrateDB table DDL.
         */
        createTableFromSchema,
        schemaName,
        tableName,
        qualifiedName,
        emitter,
    };
};
